import datetime
import re
import subprocess
from typing import List, Optional


def is_dirty() -> bool:
    output = subprocess.check_output(["git", "status", "--porcelain"])
    return bool(output.strip())


def create_branch(name: str) -> None:
    subprocess.check_call(["git", "checkout", "-b", name])


def delete_branch(name: str) -> None:
    subprocess.check_call(["git", "branch", "-D", name])


def repo_from_remote() -> str:
    output = (
        subprocess.check_output(["git", "remote", "get-url", "origin"])
        .decode("utf-8")
        .strip()
    )
    matches = re.search(r"/([^/]+/[^/]+?)(\.git)?$", output)
    if matches:
        return matches[1]
    else:
        return ""


def add_commit(path: str, message: str) -> None:
    subprocess.check_call(["git", "add", path])
    subprocess.check_call(["git", "commit", "-m", message])


def push(branch: str) -> None:
    subprocess.check_call(
        ["git", "push", "--force", "--set-upstream", "origin", branch]
    )


def checkout(ref: str) -> None:
    subprocess.check_call(["git", "checkout", ref])


def last_commit_message() -> str:
    return (
        subprocess.check_output(["git", "show", "-s", "--format=%s"])
        .decode("utf-8")
        .strip()
    )


def last_commit_files_added() -> List[str]:
    return (
        subprocess.check_output(
            ["git", "diff", "HEAD^", "HEAD", "--name-only", "--diff-filter", "A"]
        )
        .decode("utf-8")
        .strip()
        .splitlines()
    )


def commit_datetime_of_path(path: str) -> Optional[datetime.datetime]:
    try:
        date_string = (
            subprocess.check_output(
                ["git", "log", "-1", "--format=%cd", "--date=iso-strict", path]
            )
            .decode("utf-8")
            .strip()
        )
        return datetime.datetime.fromisoformat(date_string)
    except (ValueError, subprocess.CalledProcessError):
        return None
